#include "utils.h"
#include <math.h>

using namespace std;


namespace Utils{
    
    string username;
    int errors;
    string scoreFilePath;
    string codeFilePath;
    map<string, Level> levelMap;
    map<string, Line> lineMap;
    string msCypher;
    
    /////////////////////////////////////////////////////////////////////////////////////////////////////////
    
    
    void submitAnswer(string level, string word)
    {
        
        Line line;
        
        line.crc = calcrc(word);
        
        
        if(level == "000")
        {
            line.level = encryptLevel(level, line.crc);
            string t_crc = line.level + line.crc;
            line.word = encryptWord(word ,t_crc);
            lineMap[level] = line;
        }
        else
        {
            
            if(lineMap.find("000") == lineMap.end())
            {
                cout << "ERROR: username not found \n";
                return;
            }
            
            line.crc = addCRCs(lineMap["000"].crc, line.crc);
            string t_crc = encryptLevel(level, line.crc);
            t_crc += line.crc;
            line.word = encryptWord(word ,t_crc);
            
            int ilevel = stoi(level);
            int sl = ilevel%100;
            int ml = ilevel/100;
            
            for(int i = 0; i < sl; i++)
            {
                string slevel = intToString(ml * 100 + i);
                
                if(lineMap.find(slevel) != lineMap.end())
                {
                    line.crc = addCRCs(lineMap[slevel].crc, line.crc);
                }
                
            }
            
            line.level = encryptLevel(level, line.crc);
            lineMap[level] = line;
        }
        
    }
    
    ////////////////////////////////////////////////////////////////////////////////////////////////////////////
    
    int calcScores()
    {
        
        int score = 0;
        
        if(lineMap.find("000") != lineMap.end())
        {
            
            if(checkUsername())
            {
                score += 1;
            }else{
                cout << "ERROR: incorrect username \n";
                return 0;
            }
            
            int cLevel = 100;
            
            while(cLevel < 700)
            {
                
                if(lineMap.find(intToString(cLevel)) != lineMap.end())
                {
                    int subLevel = cLevel;
                    
                    while(subLevel > 0)
                    {
                        string sls = intToString(subLevel);
                        
                        if(levelMap.find(sls) != levelMap.end())
                        {
                            if(lineMap.find(sls) != lineMap.end())
                            {
                                if(checkLevelCode(subLevel, lineMap[sls].crc))
                                {
                                    score += levelMap[sls].points;
                                }
                            }
                            
                            subLevel += 1;
                            
                        }
                        else
                        {
                            subLevel = -1;
                        }
                    }
                    
                    cLevel += 100;
                    
                }
                else
                {
                    cLevel += 100;
                }
                
            }
            
        }
        
        return score;
    }
    
    bool checkUsername(){
        return calcrc(username) == lineMap["000"].crc;
    }
    
    int getFinalSubLevel(int level){
        
        int ul = level + 99;
        int finalLevel = 0;
        
        while(level < ul)
        {
            string slevel = intToString(level);
            if(lineMap.find(slevel) != lineMap.end())
            {
                finalLevel = level;
            }
            
            level++;
        }
        
        return finalLevel;
    }
    
    string getLevelWord(int level)
    {
        int sl = level%100;
        int ml = level/100;
        
        string crc = lineMap[intToString(level)].crc;
        
        for(int i = sl-1; i >= 0; i--)
        {
            
            string slevel = intToString(ml * 100 + i);
            
            if(lineMap.find(slevel) != lineMap.end())
            {
                crc = subCRCs(crc, lineMap[slevel].crc);
            }
            
        }
        
        string ls = intToString(level);
        Line l = lineMap[ls];
        string tl = encryptLevel(ls, crc);
        tl += crc;
        return decryptWord(l.word, tl);
        
    }
    
    bool checkLevelCode(int level, string crc)
    {
        int sl = level%100;
        int ml = level/100;
        
        for(int i = sl-1; i >= 0; i--)
        {
            
            string slevel = intToString(ml * 100 + i);
            
            if(lineMap.find(slevel) != lineMap.end())
            {
                crc = subCRCs(crc, lineMap[slevel].crc);
            }
            
        }
        
        string ls = intToString(level);
        Line l = lineMap[intToString(level)];
        string tl = encryptLevel(ls, crc);
        tl += crc;
        string word = decryptWord(l.word, tl);
        string answer = calcrc(word);
        
        crc = subCRCs(crc,lineMap["000"].crc);
        
        bool isValid = false;
        if(crc == answer)
        {
            if(levelMap.find(ls) != levelMap.end())
            {
                isValid = (levelMap[ls].word == word);
            }
        }
        
        return isValid;
        
    }
    
    ///////////////////////////////////////////////////////////////////////////////////////////////
    
    void readLevelsFile(){
        levelMap.clear();
        
        std::vector<string> lines;
        
        
        int count = 0;
        
        while(count < lines.size())
        {
            string k = lines[count].substr(0, 3);
            levelMap[k].points = atoi(lines[count].substr(4,1).c_str());
            levelMap[k].word = lines[count].substr(6);
            levelMap[k].level = k;
            
            count++;
        }
    }
    
    void readLevelsFile(string fp) {
        
        ifstream codeFile;
        codeFile.open(fp);
        
        if(codeFile.is_open())
        {
            
            levelMap.clear();
            string line;
            
            while(getline (codeFile, line))
            {
                if(line.length() > 0)
                {
                    string k = line.substr(0, 3);
                    levelMap[k].points = atoi(line.substr(4,1).c_str());
                    levelMap[k].word = line.substr(6);
                    levelMap[k].level = k;
                }
            }
            
            
            codeFile.close();
            
        }else{
            
            cout << "WARNING: Path to code file is incorrect \n";
        }
        
        
        
    }
    
    //////////////////////////////////////////////////////////////////////////////////////////////
    
    bool checkFilePath(string fp){
        
        
        ifstream scoreFile;
        scoreFile.open(fp);
        if(scoreFile.is_open())
        {
            scoreFile.close();
            return true;
        }else{
            
            //cout << "ERROR: Your file path is incorrect ! \n";
            //pause(1);
            return false;
        }
        
    }
    
    void readScoreFile(string fp) {
        
        
        ifstream scoreFile;
        scoreFile.open(fp);
        
        if(scoreFile.is_open())
        {
            
            lineMap.clear();
            string str;
            
            while(getline (scoreFile, str))
            {
                if(str.length() > 0)
                {
                    string ld = decryptLevel(str);
                    Line line;
                    line.level = str.substr (0,3);
                    line.crc = str.substr(3,5);
                    line.word = str.substr(8);
                    lineMap[ld] = line;
                }
            }
            
            
            scoreFile.close();
            
            
            
        }
        
        if(lineMap.find("000") != lineMap.end())
        {
            string t_crc = lineMap["000"].level + lineMap["000"].crc;
            username = decryptWord(lineMap["000"].word , t_crc);
        }
        
    }
    
    void writeScoreFile(string fp){
        
        ofstream scoreFile;
        scoreFile.open(fp, std::ofstream::out | std::ofstream::trunc);
        
        if(scoreFile.is_open())
        {
            
            map<string, Line> :: iterator it = lineMap.begin();
            while(it != lineMap.end()){
                
                string line = "";
                line = it->second.level + it->second.crc + it->second.word;
                line += "\n";
                scoreFile << line;
                it++;
            }
            
            scoreFile.close();
            
        }
    }
    
    void setUsername(string s)
    {
        username = s;
        submitAnswer("000", username);
    }
    
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////
    
    
    string decryptWord(string word, string crc)
    {
        
        int mul = atoi(crc.c_str());
        
        vector<int> cypher;
        
        for(int i = 0; i < crc.length(); i++)
        {
            char c = crc[i];
            cypher.push_back(((c-'0' + 1) * mul)%13 + 1);
        }
        
        for(int i = 0; i < word.length(); i++)
        {
            word[i] += cypher[i%cypher.size()];
        }
        
        return word;
    }
    
    string encryptWord(string word, string crc)
    {
        
        int mul = atoi(crc.c_str());
        
        vector<int> cypher;
        
        for(int i = 0; i < crc.length(); i++)
        {
            char c = crc[i];
            cypher.push_back(((c - '0' + 1) * mul)%13 + 1);
        }
        
        
        
        for(int i = 0; i < word.length(); i++)
        {
            word[i] -= cypher[i%cypher.size()];
        }
        
        return word;
    }
    
    
    
    string decryptLevel(string str)
    {
        string level = str.substr(0,3);
        string ansCrc = str.substr(3, 3);
        level = addCRCs(level, ansCrc);
        return level;
    }
    
    
    
    string encryptLevel(string level, string crc)
    {
        
        if(level.length() <= crc.length())
        {
            string sub = crc.substr(0, level.length());
            
            string cryptLevel = subCRCs(level, sub);
            
            return cryptLevel;
        }
        else
        {
            
            cout << "ERROR: key is too long \n";
            return "";
        }
        
    }
    
    /////////////////////////////////////////////////////////////////////////////////////////////////////////////
    
    string calcrc(const char * ptr)
    {
        
        
        int count = strlen(ptr);
        unsigned short crc;
        
        crc = 0;
        
        //
        for(int j = 0; j < count; j++)
        {
            
            int p = *ptr;
            p = p << 8;
            crc = crc ^ p;
            
            for(int i = 0; i < 8; i++)
            {
                if (crc & 0x8000)
                    crc = crc << 1 ^ 0x1021;
                else
                    crc = crc << 1;
            }
            
            ++ptr;
        }
        
        string str = intToString(crc);
        
        while(str.length() < 5)
        {
            str = "0" + str;
        }
        
        return str;
    }
    
    
    
    string calcrc(string st)
    {
        
        char * ptr = new char[st.length() + 1];
        strcpy(ptr, st.c_str());
        char * savedPtr = ptr;
        
        string crc = calcrc(ptr);
        
        delete []savedPtr;
        return crc;
    }
    
    string addCRCs(string a, string b)
    {
        
        if(a.length() == b.length())
        {
            
            string str(a.length(), ' ');
            
            for(int i = 0; i < a.length(); i++)
            {
                char ac = a[i];
                char bc = b[i];
                int ai = ac - '0';
                int bi = bc - '0';
                str[i] = '0' + (ai + bi)%10;
            }
            return str;
        }
        else
        {
            return "";
        }
        
    }
    
    string subCRCs(string a, string b)
    {
        
        if(a.length() == b.length())
        {
            
            string str(a.length(), ' ');
            
            for(int i = 0; i < a.length(); i++)
            {
                char ac = a[i];
                char bc = b[i];
                int ai = ac - '0';
                int bi = bc - '0';
                int r = (ai - bi);
                if(r < 0) r += 10;
                str[i] = '0' + r;
            }
            return str;
        }
        else
        {
            return "";
        }
        
    }
    
    
    ////////////////////////////////////////////////////////////////////////////////////////////////////
    
    void split(const string& s, char c, vector<string>& v)
    {
        string::size_type i = 0;
        string::size_type j = s.find(c);
        
        if(j == string::npos){
            v.push_back(s);
        }
        
        while (j != string::npos)
        {
            v.push_back(s.substr(i, j-i));
            i = ++j;
            j = s.find(c, j);
            
            if (j == string::npos)
            {
                v.push_back(s.substr(i, s.length()));
            }
        }
    }
    
    string intToString(int a){
        
        stringstream ss;
        ss << a;
        return ss.str();
    }
    
    ////////////////////////////////////////////////////////////////////////////
    
    void millisleep(int ms) {
#ifdef WIN32
        Sleep(ms);
#else
        struct timespec ts = {0}, rem;
        ts.tv_sec = ms / 1000;
        ts.tv_nsec = (ms % 1000) * 1000000L;
        while(nanosleep(&ts, &rem)) {
            ts.tv_sec = rem.tv_sec;
            ts.tv_nsec = rem.tv_nsec;
        }
#endif
    }
    
    void pause(int n) {
        for(int i = 0; i < n; i++)
        {
            printf( "... \n" );
            millisleep(700);
        }
    }
    
    void printVerticalDots(int num, int ms) {
        for(int i = 0; i < num; i++)
        {
            printf(".\n");
            millisleep(ms);
        }
    }
    ///////////////////////////////////////////////////////////////////////////////
    
    string encryptScript(string s)
    {
        string output = "";
        string allChars = "abcdefghijkalmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ':@/.?";
        
        
        for(int j=0; j < s.size(); j++){
            
            if(s[j] == ' '){
                
                output+= " ";
                
            }else{
                
                for(int i =0; i <   allChars.size(); i++)
                {
                    if(s[j] == allChars[i]){
                        output += msCypher[i];
                        break;
                    }
                }
            }
            
            
        }
        
        return output;
    }
    
    string decryptScript(string s)
    {
        string output = "";
        string allChars = "abcdefghijkalmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ':@/.?";
        
        for(int j=0; j < s.size(); j++){
            
            if(s[j] == ' '){
                
                output+= " ";
                
            }else{
                
                for(int i =0; i <   msCypher.size(); i++)
                {
                    if(s[j] == msCypher[i]){
                        output += allChars[i];
                        break;
                    }
                }
            }
            
            
        }
        
        return output;
        
    }
    
    string setScriptCypher(string s)
    {
        msCypher = s;
    }
    
}
